#include "mem_pool.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <tmc/alloc.h>
#include <tmc/mspace.h>
#include <stdint.h>
#define LIMIT(x) (((addr_type)1)<<x) 
#define PAGEMAP_ENTRY 8
#define GET_BIT(X,Y) (X & ((uint64_t)1<<Y)) >> Y
//! Entry is lower 54 bits(if present)
#define GET_PFN(X) X & 0x7FFFFFFFFFFFFF
const int __endian_bit = 1;
#define is_bigendian() ( (*(char*)&__endian_bit) == 0 )
# define STRIPED_MEMORY 1
tmc_alloc_t alloc = TMC_ALLOC_INIT;
typedef struct
{
    unsigned int colbits:10;
    unsigned int bankbits:2;
    unsigned int rowbits:13;
    unsigned int chip:2; 
}Address;

//! size in bytes
# define SIZE 1494967295
uint64_t read_val, file_offset;
FILE * f;
int page_bits_mask            = 0x0;
unsigned int page_bits        = 0;
char *pool_base_addr          = 0x0;
unsigned int *mappings        = 0x0;
unsigned int size_elmts       = 0x0;
unsigned int size_mem         = 0x0;
unsigned int avail_size_mem   = 0x0;
addr_type granularity         = 0; 
short l_colour                = -1; 
void updatePageSize()
{
    unsigned int pg_size = getpagesize();
    # ifdef STRIPED_MEMORY
    //granularity          = 8192; 
    granularity          = 4096; 
    # else
    granularity          = getpagesize();
    # endif
    while(pg_size != 1)
    {
        pg_size = pg_size >> 1;
        ++page_bits;
    }
    page_bits_mask = (1 << page_bits)-1;
    printf("\n Pg msk : %d %d %d %ld %ld \n",page_bits_mask,page_bits,getpagesize(),tmc_alloc_get_available_nodes(),4494967296);//sizeof(addr_type));
}

int translate_vaddr(char * path_buf, addr_type virt_addr , addr_type *pa)
{
   int i, c, pid, status;
   printf("Big endian? %d\n", is_bigendian());
   f = fopen(path_buf, "rb");
   if(!f){
      printf("Error! Cannot open %s\n", path_buf);
      return -1;
   }
   
   //Shifting by virt-addr-offset number of bytes
   //and multiplying by the size of an address (the size of an entry in pagemap file)
   //file_offset = virt_addr / getpagesize() * PAGEMAP_ENTRY;
   file_offset = (virt_addr>>page_bits) * PAGEMAP_ENTRY;
   printf("Vaddr: 0x%lx, Page_size: %d, Entry_size: %d\n", virt_addr, getpagesize(), PAGEMAP_ENTRY);
   printf("Reading %s at 0x%llx 0x%x \n", path_buf, (unsigned long long) file_offset, virt_addr & page_bits_mask);
   status = fseek(f, file_offset, SEEK_SET);
   if(status){
      perror("Failed to do fseek!");
      return -1;
   }
   errno = 0;
   read_val = 0;
   unsigned char c_buf[PAGEMAP_ENTRY];
   for(i=0; i < PAGEMAP_ENTRY; i++){
      c = getc(f);
      if(c==EOF){
         printf("\nReached end of the file\n");
         return 0;
      }
      if(is_bigendian())
           c_buf[i] = c;
      else
           c_buf[PAGEMAP_ENTRY - i - 1] = c;
      printf("[%d]0x%x ", i, c);
   }
   for(i=0; i < PAGEMAP_ENTRY; i++){
      //printf("%d ",c_buf[i]);
      read_val = (read_val << 8) + c_buf[i];
   }
   printf("\n");
   printf("Result: 0x%llx\n", (unsigned long long) read_val);
   *pa = 0x0; 
   //if(GET_BIT(read_val, 63))
   if(GET_BIT(read_val, 63))
   {
      printf("PFN: 0x%llx\n",(unsigned long long) GET_PFN(read_val));
      *pa = GET_PFN(read_val);
      *pa = *pa <<16;
   }
   else
      printf("Page not present\n");
   if(GET_BIT(read_val, 62))
      printf("Page swapped\n");
   fclose(f);
   return 0;
}
/*{
    int i, c, pid, status;
    printf("Big endian? %d\n", is_bigendian());
    //! Open the mapping file
    f = fopen(path_buf, "rb");
    if(!f)
    {
        printf("Error! Cannot open %s\n", path_buf);
        return -1;
    }
   
    //Shifting by virt-addr-offset number of bytes
    //and multiplying by the size of an address (the size of an entry in pagemap file)
    //file_offset = virt_addr / getpagesize() * PTE;
    file_offset = (virt_addr>>page_bits) * PTE;
    printf("Vaddr: 0x%lx, Page_size: %d, Entry_size: %d\n", virt_addr, getpagesize(), PTE);
    printf("Reading %s at 0x%llx 0x%x \n", path_buf, (unsigned long long) file_offset, virt_addr & page_bits_mask);
    status = fseek(f, file_offset, SEEK_SET);
    if(status)
    {
        perror("Failed to do fseek!");
        return -1;
    }
    errno = 0;
    read_val = 0;
    unsigned char c_buf[PTE];
    for(i=0; i < PTE; i++)
    {
        c = getc(f);
        if(c==EOF)
        {
            printf("\nReached end of the file\n");
            return 0;
        }
        if(is_bigendian())
            c_buf[i] = c;
        else
            c_buf[PTE - i - 1] = c;
        printf("[%d]0x%x ", i, c);
    }
    for(i=0; i < PTE; i++)
    {
        read_val = (read_val << 8) + c_buf[i];
    }
    printf("\n");
    printf("Result: 0x%llx\n", (unsigned long long) read_val);
    *pa = 0x0; 
     //! Figure out if page is actually mapped to physical memory or not
    if(GET_BIT(read_val, 63))
    {
        *pa = GET_PFRAME(read_val);
        printf("PFN: 0x%llx\n",(unsigned long long) *pa);
        *pa = *pa <<16;
    }
    else
        printf("Page not present\n");
    if(GET_BIT(read_val, 62))
       printf("Page swapped\n");
    fclose(f);
    return 0;
}*/

short figure_out_controller(addr_type virt_addr)
{
    # ifdef STRIPED_MEMORY
    /*addr_type offset  = virt_addr & page_bits_mask;
    short colour = offset/8192;
    colour = (colour > 3 ) ? colour-4 : colour;*/
    addr_type offset = virt_addr >> 12;
    short colour = offset & 0x3;
    printf("\n %x %x %d ",virt_addr,offset,colour);
    return colour; 
    # else
    addr_type pa          = 0x0;
    char *x               = (char *)virt_addr; 
    int pid               = getpid(); 
    char path_buf [0x100] = {};
    //! To bring the page into memory if its not there 
    volatile char ct = *x;
    *x = ct;
    sprintf(path_buf, "/proc/%u/pagemap", pid);
    translate_vaddr(path_buf,virt_addr,&pa);
    //pa = pa >> 25;
    pa = pa >> 34;
    pa = pa & 0b11;
    return (short)(pa);
    # endif
    return -1;
}

void updateMappings()
{
    if(mappings)
    {
        addr_type i = 0;
        for(;i<size_elmts;i++)
            mappings[i] = figure_out_controller((addr_type)(pool_base_addr+(i*granularity)));
    }
}

int init_pool(addr_type size)
{
    updatePageSize();
    size    = ((size-1)/granularity)+1;
    size   *= granularity;
    //! This will be a shared mapping
    tmc_alloc_set_shared(&alloc);
    //! Disable caching
    tmc_alloc_set_caching(&alloc, MAP_CACHE_NONE); 
    pool_base_addr = (char*)tmc_alloc_map(&alloc,size);
    printf("\n Pool Intiated .... 0x%x\n",pool_base_addr);
    if(pool_base_addr)
    {
        size_mem         = size;
        avail_size_mem   = size;
        size_elmts       = (size/granularity)+1;
        mappings         = (unsigned int *)malloc(size_elmts*sizeof(unsigned int));
        updateMappings();
        return 1;
    }
    return 0;
}

char* get_memory(addr_type size)
{
    //! Limits
    //! round up memory to page size boundary
    size    = ((size-1)/granularity)+1;
    size   *= granularity;
    //printf("\n @@@@@@@@@@@@@@@@@@@@@ %ld %ld %ld ",size,avail_size_mem,granularity);
    if(size > 0 && size <= avail_size_mem)
    {
        char *ptr = getmemory(size);
        avail_size_mem -= size;
        //printf("\n AM %ld \n",avail_size_mem); 
        return ptr;        
    }
    return 0x0;      
}

char* getmemory(addr_type size)
{
    addr_type x = 0;
    if(size <= granularity)
    {
        //!
        char flag  = 0;
        char flag1 = 0;
        //!
        addr_type last_free_slot;
        l_colour += 1;
        l_colour &= 0x3;
        //printf("\n Color : %d ",l_colour);
        for(;x<size_elmts;x += 1)
        {
            //! Slot available
            if(mappings[x] <= 3)
            {
                //! Match found
                if(mappings[x] == l_colour)
                {
                    flag = 1;
                    break;
                }
                else
                {
                    last_free_slot = x;
                    flag1 = 1;
                }
            }
        }
        //! We found a req. slot matching criteria
        if(flag == 1)
        {
            //!
            l_colour       = mappings[x];
            char *addr     = (char*)(pool_base_addr + (x*granularity));
            //printf("\n Allotted %d %x ",mappings[x],(addr_type)addr);
            mappings[x]    = (addr_type)addr;
            return addr;
        }
        if( x == size_elmts && flag1 == 1)
        {
            l_colour       = mappings[last_free_slot];
            char *addr     = (char*)(pool_base_addr + (last_free_slot*granularity));
            //printf("\n Allotted %d %x ",mappings[last_free_slot],(addr_type)addr);
            mappings[last_free_slot]    = (addr_type)addr;
            return addr;

        }
    }
    else
    {
        x = 0;
        char flag = 0;
        addr_type y = x,k=0;
        size /= granularity;
        for(;x<size_elmts;x++)
        {
            ++k;
            if(mappings[x] <= 3)
            {
                for(y=x+1;y<size_elmts;y++)
                {
                    if(mappings[y] <= 3)
                    {
                        ++k;
                        if(k==size)
                        {
                            flag = 1;
                            break;
                        }         
                    }
                    else
                    {
                        break;
                    }
                }
                if(flag == 1)
                {
                    //!
                    break;
                }
                else
                {
                    k=0;
                } 
            }
        }
        if(flag == 1)
        {
            l_colour       = mappings[x+size-1];
            char *addr     = (char*)(pool_base_addr + (x*granularity)); 
            for(y=x;y<(x+size);y++)
            {
                mappings[y] = (addr_type)addr; 
            }
            return addr;
        }
    }
    //!
    return 0x0;
}

extern char* get_memory_c(addr_type size, short colour)
{
    //! Limits
    //! round up memory to page size boundary
    size    = ((size-1)/granularity)+1;
    size   *= granularity;
    //printf("\n @@@@@@@@@@@@@@@@@@@@@ %ld %ld %ld ",size,avail_size_mem,granularity);
    if(size > 0 && size <= avail_size_mem)
    {
        char *ptr = getmemoryc(size,colour);
        avail_size_mem -= size;
        //printf("\n AM %ld \n",avail_size_mem); 
        return ptr;        
    }
    return 0x0;      
}

char* getmemoryc(addr_type size, short colour)
{
    addr_type x = 0;
    if(size <= granularity)
    {
        //!
        char flag  = 0;
        char flag1 = 0;
        //!
        addr_type last_free_slot;
        //printf("\n Color : %d ",l_colour);
        for(;x<size_elmts;x += 1)
        {
            //! Slot available
            if(mappings[x] <= 3)
            {
                //! Match found
                if(mappings[x] == colour)
                {
                    flag = 1;
                    break;
                }
                else
                {
                    last_free_slot = x;
                    flag1 = 1;
                }
            }
        }
        //! We found a req. slot matching criteria
        if(flag == 1)
        {
            //!
            l_colour       = mappings[x];
            char *addr     = (char*)(pool_base_addr + (x*granularity));
            //printf("\n Allotted %d %x ",mappings[x],(addr_type)addr);
            mappings[x]    = (addr_type)addr;
            return addr;
        }
        if( x == size_elmts && flag1 == 1)
        {
            l_colour       = mappings[last_free_slot];
            char *addr     = (char*)(pool_base_addr + (last_free_slot*granularity));
            //printf("\n Allotted %d %x ",mappings[last_free_slot],(addr_type)addr);
            mappings[last_free_slot]    = (addr_type)addr;
            return addr;

        }
    }
    else
    {
        x = 0;
        char flag = 0;
        addr_type y = x,k=0;
        size /= granularity;
        for(;x<size_elmts;x++)
        {
            ++k;
            if(mappings[x] <= 3)
            {
                for(y=x+1;y<size_elmts;y++)
                {
                    if(mappings[y] <= 3)
                    {
                        ++k;
                        if(k==size)
                        {
                            flag = 1;
                            break;
                        }         
                    }
                    else
                    {
                        break;
                    }
                }
                if(flag == 1)
                {
                    //!
                    break;
                }
                else
                {
                    k=0;
                } 
            }
        }
        if(flag == 1)
        {
            l_colour       = mappings[x+size-1];
            char *addr     = (char*)(pool_base_addr + (x*granularity)); 
            for(y=x;y<(x+size);y++)
            {
                mappings[y] = (addr_type)addr; 
            }
            return addr;
        }
    }
    //!
    return 0x0;
}

void release_pool()
{
    if(pool_base_addr != 0x0)
    {
        tmc_alloc_unmap(pool_base_addr,size_mem);
        pool_base_addr = 0x0;
    }
}
