#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <tmc/alloc.h>

#include <arch/cycle.h>
#include <arch/spr.h>

#include <tmc/cpus.h>
#include <tmc/sync.h>
#include <tmc/task.h>
#include <tmc/alloc.h>
#include <tmc/mspace.h>
#include "mem_pool.h"
# define POOL 1
#define NUM_THREADS 4
pthread_t threads[NUM_THREADS];
cpu_set_t cpus;
pthread_mutex_t      mutex = PTHREAD_MUTEX_INITIALIZER;
tmc_alloc_t alloc1   = TMC_ALLOC_INIT;
//const int ITERATIONS = 10000000;
const int ITERATIONS = 10;
addr_type size       = 1966080000;//LIMIT(34);  
//addr_type size       = 3216637952;//LIMIT(34);  
//addr_type size       = 1494800000;//LIMIT(34);  
//addr_type size       = 536838144;//LIMIT(34);  
addr_type chunk_size = 8192; 
char **test_array ;
addr_type num_elmts  = 30000;
// Barrier objects for synchronizing 2 threads or all threads.
tmc_sync_barrier_t barrier_all = TMC_SYNC_BARRIER_INIT(NUM_THREADS);


// Measure the average load latency to a particular memory address.
// This routine will stride through the specified memory buffer,
// performing a load every 'stride' bytes and wrapping after 'bound'
// bytes.  'bound' must be a power of two.
//
uint64_t measure_load_latency(int rank)
{
  uint64_t start = get_cycle_count();
  addr_type i = (addr_type)rank;
  for (int k = 0; k < ITERATIONS; k++)
  {
      // Load into a useless SPR so that the compiler won't drop the
      // load or do any interesting scheduling.  We chose this SPR
      // because it accessible to userspace and has a single-cycle
      // access time.
      i = rank;
      addr_type j;
      for(j=0;j<chunk_size;j++)
      {
          for(;i<num_elmts;i=i+NUM_THREADS)
          {
              //! write something
              test_array[i][j] = '\0';
              // __insn_mtspr(SPR_SYSTEM_SAVE_0_0, base[offset]);
              //offset = (offset + stride) & mask;
          }
      }
  }
  uint64_t finish = get_cycle_count();
	
  // Calculate the average load latency.  We know that the compiler 
  // has to schedule the load, mtspr, and branch operation in separate
  // bundles, and we want to measure how many cycles were consumed in 
  // addition to those bundles, so we subtract three from the average
  // loop iteration time.
  uint64_t c_count = (finish - start) / ITERATIONS;
  addr_type nt_elmts = num_elmts/NUM_THREADS;
  c_count /= (nt_elmts);
  return c_count;
}


// The worker function for each thread.  We invoke each memory example
// function in turn and then exit.
//
void* thread_func(void* arg)
{
  int *temp = (int*)arg;
  int rank  = *temp;
  int cpu   = 0;
  switch(rank)
  {
      case 0:
         cpu = tmc_cpus_grid_tile_to_cpu(6,6);
         break;
      case 1:
         cpu = tmc_cpus_grid_tile_to_cpu(1,6);
         break;
      case 2:
         cpu = tmc_cpus_grid_tile_to_cpu(6,1);
         break;
      case 3:
         cpu = tmc_cpus_grid_tile_to_cpu(1,1);
         break;
  }
  //printf("\n Thread ID %d \n",rank); 
  // Bind this thread to the rank'th core in the cpu set.
  if (tmc_cpus_set_my_cpu(tmc_cpus_find_nth_cpu(&cpus, cpu)) < 0)
    tmc_task_die("tmc_cpus_set_my_cpu() failed.");
  # ifndef POOL 
  addr_type  i = rank;
  pthread_mutex_lock(&mutex);
  for(;i<num_elmts;i+=NUM_THREADS)
  //for(;i<10;i++)
  {
      test_array[i] = (char*)tmc_alloc_map(&alloc1,chunk_size);
      //test_array[i] = (char*)(&tmp[i*chunk_size]);
      //test_array[i] = (char*)get_memory(chunk_size);
      //printf("\n %ld %x ",i,(unsigned int)(test_array[i]));
  }
  pthread_mutex_unlock(&mutex);
  # endif
  tmc_sync_barrier_wait(&barrier_all);
  uint64_t cycles_consumed  =  measure_load_latency(rank);
  printf("\n Cycles Consumed : index %d cycles %lld",rank,cycles_consumed); 
  printf("\n Assigned CPU %d %d \n",rank,tmc_cpus_find_nth_cpu(&cpus, cpu));
  # ifndef POOL 
  pthread_mutex_lock(&mutex);
  for(i=rank;i<num_elmts;i+=NUM_THREADS)
  {
      tmc_alloc_unmap(test_array[i],chunk_size);
  }
  pthread_mutex_unlock(&mutex);
  # endif
  return NULL;
}

int arr[NUM_THREADS];
// This method creates (count - 1) child threads and calls 'func' in
// each thread, passing the rank to each thread.  The caller is
// assigned the lowest rank.
//
void
run_threads(int count, void*(*func)(void*))
{
    unsigned int i = 0; 
    arr[i] = i;
    int status = pthread_create(&threads[i], NULL, func, (void*)&arr[i]);
    if (status != 0)
    {
        printf("\n Err %d ",status);
        tmc_task_die("pthread_create() failed.");
    }
    printf("\n id %d %d ",i,count);
    ++i;
    arr[i] = i;
    status = pthread_create(&threads[i], NULL, func, (void*)&arr[i]);
    if (status != 0)
    {
        printf("\n Err %d ",status);
        tmc_task_die("pthread_create() failed.");
    }
    printf("\n id %d %d ",i,count);
    ++i;
    arr[i] = i;
    status = pthread_create(&threads[i], NULL, func, (void*)&arr[i]);
    if (status != 0)
    {
        printf("\n Err %d ",status);
        tmc_task_die("pthread_create() failed.");
    }
    printf("\n id %d %d ",i,count);
    ++i;
    arr[i] = i;
    status = pthread_create(&threads[i], NULL, func, (void*)&arr[i]);
    if (status != 0)
    {
        printf("\n Err %d ",status);
        tmc_task_die("pthread_create() failed.");
    }
    printf("\n id %d %d ",i,count);
}


// Wait for all threads to finish.
//
void
finish_threads(int count)
{
    if (pthread_join(threads[0], NULL) != 0)
      tmc_task_die("pthread_join() failed.");
    if (pthread_join(threads[1], NULL) != 0)
      tmc_task_die("pthread_join() failed.");
    if (pthread_join(threads[2], NULL) != 0)
      tmc_task_die("pthread_join() failed.");
    if (pthread_join(threads[3], NULL) != 0)
      tmc_task_die("pthread_join() failed.");
}


int
main(int argc, char**argv)
{
  pthread_mutex_init(&mutex,NULL);
  // Make sure we have enough cpus.
  if (tmc_cpus_get_my_affinity(&cpus) != 0)
    tmc_task_die("tmc_cpus_get_my_affinity() failed.");
  if (tmc_cpus_count(&cpus) < (NUM_THREADS))
    tmc_task_die("Insufficient cpus available.");
   # ifndef POOL 
   size = num_elmts * getpagesize();
   # else
   size = num_elmts * chunk_size;
   # endif
   printf("\n NE %ld %ld \n",num_elmts,size); 
   tmc_alloc_set_shared(&alloc1);
   tmc_alloc_set_caching(&alloc1, MAP_CACHE_NONE); 
   test_array = (char**)malloc(num_elmts);
   //char *tmp = (char*)calloc(size,sizeof(char));
   # ifdef POOL
   init_pool(num_elmts*chunk_size); 
   addr_type i = 0;
   for(;i<num_elmts;i++)
   {
       test_array[i] = (char*)get_memory(chunk_size);
       //printf("\n %ld %x ",i,(unsigned int)(test_array[i]));
   }
   # endif
   // Call the main thread function on each cpu, then wait for them all
   // to exit.
   run_threads(NUM_THREADS, thread_func);
   finish_threads(NUM_THREADS);
   # ifdef POOL
   release_pool();
   # endif
  return 0;
}
